<?php
namespace Datec\DatecFeusermanagement\Services;

/*
 * This file is part of the TYPO3 CMS project.
 *
 * It is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, either version 2
 * of the License, or any later version.
 *
 * For the full copyright and license information, please read the
 * LICENSE.txt file that was distributed with this source code.
 *
 * The TYPO3 project - inspiring people to share!
 */

use TYPO3\CMS\Core\Utility\GeneralUtility;

/**
 * Authentication services class
 * 
 * PR: customized for login with remote interface of KIS
 */
class RemoteAuthenticationService extends \TYPO3\CMS\Sv\AuthenticationService {
	
	/**
	 * @var array
	 */
	protected $settings;
	
	/**
	 * Initialize authentication service
	 * Saves remote authentication hash in session.
	 *
	 * @param string $mode Subtype of the service which is used to call the service.
	 * @param array $loginData Submitted login form data
	 * @param array $authInfo Information array. Holds submitted form data etc.
	 * @param AbstractUserAuthentication $pObj Parent object
	 * @return void
	 */
	public function initAuth($mode, &$loginData, $authInfo, $pObj) {
		parent::initAuth($mode, $loginData, $authInfo, $pObj);
		
		if (isset($loginData['uident'])) {
			// build remote authentification hash
			$userRemoteAuthHash = base64_encode($loginData['uname'].':'.$loginData['uident']);
			// store for this session
			$this->pObj->setAndSaveSessionData('userRemoteAuthHash', $userRemoteAuthHash);
			$loginData['userRemoteAuthHash'] = $userRemoteAuthHash;
			$this->login['userRemoteAuthHash'] = $userRemoteAuthHash;
		}
	}

	/**
	 * Find a user (eg. look up the user record in database when a login is sent)
	 *
	 * @return mixed User array or FALSE
	 */
	public function getUser() {
		if ($this->login['status'] !== 'login') {
			return FALSE;
		}
		if (!$this->login['uident']) {
			// Failed Login attempt (no password given)
			$this->writelog(255, 3, 3, 2, 'Login-attempt from %s (%s) for username \'%s\' with an empty password!', array(
				$this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']
			));
			GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), for username \'%s\' with an empty password!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']), 'Core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
			return FALSE;
		}

		if (isset($this->login['userRemoteAuthHash'])) {
			$applicationUser = $this->getApplicationUserRemote($this->login['userRemoteAuthHash']);	
			if (!is_object($applicationUser)) {
				$this->writelog(255, 3, 3, 2, 'Login-attempt from %s (%s), username \'%s\' APPLICATIONUSER not found!!', array($this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']));
				GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), username \'%s\' APPLICATIONUSER not found!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']), 'core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
			} else {	
				$person = $this->getPersonRemote($this->login['userRemoteAuthHash'], $applicationUser->entityTypeObj->person->id);
				if (!is_object($person)) {
					$this->writelog(255, 3, 3, 2, 'Login-attempt from %s (%s), username \'%s\' PERSON not found!!', array($this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']));
					GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), username \'%s\' PERSON not found!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']), 'core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
				} else {
					// user exists
					$table = 'fe_users';
					$usergroups = array();
					$confArray = unserialize($GLOBALS['TYPO3_CONF_VARS']['EXT']['extConf']['datec_feusermanagement']);					
					// cleanup before lookup
					$this->deleteOldUsers($applicationUser->entityTypeObj->id, $this->login['uname']);
					// lookup user
					$user = $this->fetchUserRecord($this->login['uname']);
					if (!is_array($user)) {						
						// insert person as user
						if (count($person->entityTypeObj->roles)) {
							// TODO: get T3 groups for roles
						} else {
							$usergroups[] = $confArray['defaultFeUsergroup'];
						}
						$time = time();
						$insertArr = array(
							'username' => $this->login['uname'],
							'usergroup' => implode(',', $usergroups),	
							'tx_datec_kis_user_id' => $applicationUser->entityTypeObj->id,
							'tx_datec_kis_person_id' => $person->entityTypeObj->id,
							'last_name' => $person->entityTypeObj->lastName,
							'first_name' => $person->entityTypeObj->firstName,
							'name' => $person->entityTypeObj->firstName.' '.$person->entityTypeObj->lastName,
							'tx_extbase_type' => 'Tx_Extbase_Domain_Model_FrontendUser',
							'pid' => $confArray['storagePid'],
							'crdate' => $time,			
							'tstamp' => $time								
						);
						$res = $GLOBALS['TYPO3_DB']->exec_INSERTquery($table, $insertArr);
						if (!$res) {
							GeneralUtility::sysLog(sprintf('Could not create user in table \'%s\', affected fields %s', $table, serialize($insertArr)), 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
							return FALSE;
						}
					} else {
						// update user with person
						if (count($person->entityTypeObj->roles)) {
							// TODO: get T3 groups for roles
						} else if (count($user['usergroup'])) {
							foreach (explode(',', $user['usergroup']) as $usergroup) {
								$usergroups[] = $usergroup;
							}
						} else {
							$usergroups[] = $confArray['defaultFeUsergroup'];
						}
						$updateArr = array(
							'username' => $this->login['uname'],
							'usergroup' => implode(',', $usergroups),	
							'tx_datec_kis_user_id' => $applicationUser->entityTypeObj->id,
							'tx_datec_kis_person_id' => $person->entityTypeObj->id,
							'last_name' => $person->entityTypeObj->lastName,
							'first_name' => $person->entityTypeObj->firstName,
							'name' => $person->entityTypeObj->firstName.' '.$person->entityTypeObj->lastName,
							'tx_extbase_type' => 'Tx_Extbase_Domain_Model_FrontendUser',
							'pid' => $confArray['storagePid'],
							'tstamp' => time()
						);
						$where = 'uid=' . $user['uid'];
						$res = $GLOBALS['TYPO3_DB']->exec_UPDATEquery($table, $where, $updateArr);
						if (!$res) {
							GeneralUtility::sysLog(sprintf('Could not update user in table \'%s\', affected fields %s, where conditions \'%s\'', $table, serialize($insertArr), $where), 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
							return FALSE;
						}
						if ($this->writeDevLog) {
							GeneralUtility::devLog('User found: ' . GeneralUtility::arrayToLogString($user, array($this->db_user['userid_column'], $this->db_user['username_column'])), RemoteAuthenticationService::class);
						}
					}
					
					// reload possibly changed or added user
					$user = $this->fetchUserRecord($this->login['uname']);
					if (!is_array($user)) {
						// Failed login attempt (no username found)
						$this->writelog(255, 3, 3, 2, 'Login-attempt from %s (%s), username \'%s\' not found!!', array($this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']));
						// Logout written to log
						GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), username \'%s\' not found!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']), 'core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
					} else {
						if ($this->writeDevLog) {
							GeneralUtility::devLog('User found: ' . GeneralUtility::arrayToLogString($user, array($this->db_user['userid_column'], $this->db_user['username_column'])), RemoteAuthenticationService::class);
						}
					}
				}				
			}		
		}
			
		return $user;
	}
	
	/**
	 * Retrieves applicationUser information for user authentification hash from remote interface.
	 * 
	 * @param string $userAuthHash hash to be used in authorization header
	 * @return mixed|boolean
	 */
	private function getApplicationUserRemote($userAuthHash) {
		$confArray = unserialize($GLOBALS['TYPO3_CONF_VARS']['EXT']['extConf']['datec_feusermanagement']);
		
		$url = $confArray['remoteServiceUrl'].$confArray['remoteServices.']['auth'];		
		if (!empty($url)) {			
			$detailDataContainer = $this->buildDetailObject($confArray['contexts.']['applicationuser.']);
				
			$request = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance('TYPO3\\CMS\\Core\\Http\\HttpRequest', $url);
			$request->setMethod(\TYPO3\CMS\Core\Http\HttpRequest::METHOD_POST);
			$request->setHeader("Accept-Charset", "UTF-8");
			$request->setHeader("Content-Type", "application/json;charset=UTF-8");
			$request->setHeader("Authorization", "Basic ".$userAuthHash);
			$request->setUrl($url);
			$request->setBody(json_encode($detailDataContainer));
			
			try {
				$response = $request->send();
								
				if ($response->getStatus() === 200) {
					return json_decode($response->getBody());
				} else {
					GeneralUtility::sysLog(sprintf('Error on connection to remote interface for URL: %s, ResponseCode: %s, reason phrase: %s', $url, $response->getStatus(), $response->getReasonPhrase()) , 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
					return FALSE;
				}
			} catch (\Exception $e) {
				GeneralUtility::sysLog($e->getTraceAsString(), 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
				return FALSE;
			}			
		} else {
			return FALSE;
		}
	}
	
	/**
	 * Retrieves person information for authentificaticated application user from remote interface.
	 *
	 * @param string $userAuthHash hash to be used in authorization header
	 * @param int $personId the person identifier
	 * @return mixed|boolean
	 */
	private function getPersonRemote($userAuthHash, $personId) {
		$confArray = unserialize($GLOBALS['TYPO3_CONF_VARS']['EXT']['extConf']['datec_feusermanagement']);
		
		$url = $confArray['remoteServiceUrl'].$confArray['remoteServices.']['auth'];
		if (!empty($url)) {
			$detailDataContainer = $this->buildDetailObject($confArray['contexts.']['person.']);
			$detailDataContainer->entityTypeObjId = $personId;
				
			$request = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance('TYPO3\\CMS\\Core\\Http\\HttpRequest', $url);
			$request->setMethod(\TYPO3\CMS\Core\Http\HttpRequest::METHOD_POST);
			$request->setHeader("Accept-Charset", "UTF-8");
			$request->setHeader("Content-Type", "application/json;charset=UTF-8");
			$request->setHeader("Authorization", "Basic ".$userAuthHash);
			$request->setUrl($url);
			$request->setBody(json_encode($detailDataContainer));
			
			try {
				$response = $request->send();
	
				if ($response->getStatus() === 200) {
					return json_decode($response->getBody());
				} else {
					GeneralUtility::sysLog(sprintf('Error on connection to remote interface for URL: %s, response code: %s, reason phrase: %s', $url, $response->getStatus(), $response->getReasonPhrase()) , 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
					return FALSE;
				}
			} catch (\Exception $e) {
				GeneralUtility::sysLog($e->getTraceAsString(), 'core', GeneralUtility::SYSLOG_SEVERITY_ERROR);
				return FALSE;
			}
		} else {
			return FALSE;
		}
	}
	
	/**
	 * Converts an associtive array into an object.
	 * Respects depth of array.
	 * 
	 * @param array $array
	 * @return \stdClass|string error string as value
	 */
	private function buildDetailObject($array, $recursion = 0) {
		if ($recursion > 99) {
			return 'ERROR: recursion to deep';
		}
		$object = new \stdClass();
		
		foreach ($array as $key => $value) {
			$key = str_replace('.', '', $key); // remote configuration format for arrays
			$object->$key = is_array($value) ? $this->buildDetailObject($value, $recursion + 1) : $value;
		}
		
		return $object;
	}
	
	/**
	 * Deletes any feuser records were the username doesn't match the person id.
	 * 
	 * @param int $userId
	 * @param string $username
	 * @return boolean FALSE on error
	 */
	private function deleteOldUsers($userId, $username) {
		$user = FALSE;
		$table = 'fe_users';
		$where = 'username<>'.$GLOBALS['TYPO3_DB']->fullQuoteStr($username, $table).' AND tx_datec_kis_user_id='.$userId; // known person, wrong username
		$updateArr = array(
			'deleted' => 1
		);
		$dbres = $GLOBALS['TYPO3_DB']->exec_UPDATEquery($table, $where, $updateArr);
		if ($dbres) {
			return TRUE;
		}
		return FALSE;
	}

	/**
	 * Authenticate a user (Check various conditions for the user that might invalidate its authentication, eg. password match, domain, IP, etc.)
	 *
	 * @param array $user Data of user.
	 * @return int >= 200: User authenticated successfully.
	 *                     No more checking is needed by other auth services.
	 *             >= 100: User not authenticated; this service is not responsible.
	 *                     Other auth services will be asked.
	 *             > 0:    User authenticated successfully.
	 *                     Other auth services will still be asked.
	 *             <= 0:   Authentication failed, no more checking needed
	 *                     by other auth services.
	 */
	public function authUser(array $user) {
		$OK = 0;
		if ($this->login['uident'] && $this->login['uname'] && isset($this->login['userRemoteAuthHash'])) {			
			// Checking applicationUser found for login information:
			$applicationUser = $this->getApplicationUserRemote($this->login['userRemoteAuthHash']);
			if (is_object($applicationUser)) {
				$OK = 200;
			} else {
				// Failed login attempt (wrong password) - write that to the log!
				if ($this->writeAttemptLog) {
					$this->writelog(255, 3, 3, 1, 'Login-attempt from %s (%s), username \'%s\', password not accepted!', array($this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']));
					GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), username \'%s\', password not accepted!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $this->login['uname']), 'core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
				}
				if ($this->writeDevLog) {
					GeneralUtility::devLog('Password not accepted: ' . $this->login['uident'], RemoteAuthenticationService::class, 2);
				}
			}
			// Checking the domain (lockToDomain)
			if ($applicationUser && $user['lockToDomain'] && $user['lockToDomain'] !== $this->authInfo['HTTP_HOST']) {
				// Lock domain didn't match, so error:
				if ($this->writeAttemptLog) {
					$this->writelog(255, 3, 3, 1, 'Login-attempt from %s (%s), username \'%s\', locked domain \'%s\' did not match \'%s\'!', array($this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $user[$this->db_user['username_column']], $user['lockToDomain'], $this->authInfo['HTTP_HOST']));
					GeneralUtility::sysLog(sprintf('Login-attempt from %s (%s), username \'%s\', locked domain \'%s\' did not match \'%s\'!', $this->authInfo['REMOTE_ADDR'], $this->authInfo['REMOTE_HOST'], $user[$this->db_user['username_column']], $user['lockToDomain'], $this->authInfo['HTTP_HOST']), 'core', GeneralUtility::SYSLOG_SEVERITY_WARNING);
				}
				$OK = 0;
			}
		}
		return $OK;
	}

}
